#include <string.h>
#include <stdio.h>
#include "LRU_mech.h"
#include "plots.h"
 
#define MAX_CACHE_SIZE (1024 * 64)
#define PAGE_SIZE_BYTES (16)
 
histogram_plot   plot_mechanism;
LRU_policy       eviction_policy(MAX_CACHE_SIZE);
FILE * input_stream;
FILE * output_stream;
char buffer_string[100];
char ** x_desc;
char ** y_desc;

void put_time_in_buffer()
{
   time_t now = time(NULL); 
   strcpy(buffer_string, ctime(&now));
   if (strchr(buffer_string, '\n')) 
   {
      *strchr(buffer_string, '\n') = 0;
   };
};

void setup_plot_text(const char * executable, const char * title)
{
   plot_mechanism.set_plot_title(title);
   plot_mechanism.set_plot_title("LRU Eviction Policy");
   plot_mechanism.set_plot_subtitle("Cache Misses");
   sprintf(buffer_string, "Block Size (in bytes) = %d", PAGE_SIZE_BYTES);
   plot_mechanism.set_plot_subsubtitle(buffer_string);
   x_desc = plot_mechanism.set_x_axis_description("LRU position ");
   y_desc = plot_mechanism.set_y_axis_description(
               "Number of touchs to given position");
   put_time_in_buffer();
   strcat(buffer_string, "   ");
   if (getenv("USER"))
   {
      strcat(buffer_string, getenv("USER"));
      strcat(buffer_string, "   ");
   };
   strcat(buffer_string, executable);
   plot_mechanism.set_footnote(buffer_string);
   plot_mechanism.set_x_axis_label_style(DECIMAL_NUMBERS);
   plot_mechanism.set_y_axis_label_style(DECIMAL_NUMBERS);
   plot_mechanism.set_x_range(0, MAX_CACHE_SIZE);
   plot_mechanism.options_complete();
};
 
attempt_flag read_page_number(FILE * input_stream, UInt32 & page_number)
{
   UInt32 touch_type, memory_location;
   if ((fgets(buffer_string, 80, input_stream)) && (buffer_string[1]))
   {
      sscanf(buffer_string, "%x %x", &touch_type, & memory_location);
      page_number = memory_location/PAGE_SIZE_BYTES;
      return SUCCESS;
   }
   else
   {
      return FAILURE;
   };
}; 

int main(int argument_count, const char ** argument_vector)
{
   put_time_in_buffer(); fprintf(stdout, "Start time: %s\n", buffer_string);
   /* Initialize */
   if (argument_count != 2)
   {
      fprintf(stderr, "File name required as first and only argument\n");
      fprintf(stderr, "Aborting\n");
      exit(1);
   };
   const char * executable = argument_vector[0];
   const char * filename = argument_vector[1];
   input_stream = stdin;
#if 0
   if ((input_stream = fopen(filename, "r")) == NULL)
   {
      fprintf(stderr, "Error opening file %s for input.\n", filename);
      fprintf(stderr, "Aborting\n");
      exit(1);
   };
#endif
   setup_plot_text(executable, filename);

   /* Process */
   UInt32 page_number;
   UInt32 position;
   while (read_page_number(input_stream, page_number) == SUCCESS)
   {
      eviction_policy.record_touch(page_number, position);
      plot_mechanism.point_acculmulate(position);
   };
   eviction_policy.kill_and_test_for_corruption();
 
   /* Output */
   strcpy(buffer_string, filename);
   strcat(buffer_string,".hist1.ps");

   if ((output_stream = fopen(buffer_string, "w")) == NULL)
   {
      fprintf(stderr, "Error opening file %s for output.\n", filename);
      fprintf(stderr, "Aborting\n");
      exit(1);
   };
   plot_mechanism.print_plot(output_stream);
   fclose(output_stream);

   /* Alter minor things and output again */
   strcpy(buffer_string, filename);
   strcat(buffer_string,".hist2.ps");
   if ((output_stream = fopen(buffer_string, "w")) == NULL)
   {
      fprintf(stderr, "Error opening file %s for output.\n", filename);
      fprintf(stderr, "Aborting\n");
      exit(1);
   };
   plot_mechanism.set_histogram_type(LEFT_INTEGRATED_HIST);
   free(*y_desc); 
   *y_desc = frail_strdup("Number of memory faults");
   free(*x_desc); 
   *x_desc = frail_strdup("Cache Size");
   plot_mechanism.print_plot(output_stream);
   fclose(output_stream);
   plot_mechanism.kill();
   put_time_in_buffer();
   fprintf(stdout, "End time: %s\n", buffer_string);
};
 
 

